/*
 * General functionality supporting file paging.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "file_pager.h"
#include "file_paging.h"

#include <systypes/fcntl.h>



FilePaging::FilePaging(Pages *pages)
: _pages(pages)
{
}



/* Return any registered page mapper for the given 'fileid' or NULL if no such
   mapper is registered. */

PageMapper *FilePaging::get(fileid_t fileid)
{
    FileMapping::iterator entry = _mappers.find(fileid);
    PageMapper *mapper;

    if (entry == _mappers.end())
        mapper = NULL;
    else
        mapper = entry->second;

    return mapper;
}

/* Remove a notification manager for the given 'fileid'. */

void FilePaging::remove(fileid_t fileid, FileNotification *notifier)
{
    _notifiers.erase(fileid);
    delete notifier;
}

/* Remove a page mapper and its resources for the given 'fileid'. */

void FilePaging::remove(fileid_t fileid, PageMapper *mapper)
{
    _mappers.erase(fileid);
    delete mapper->accessor();
    delete mapper;
}

/* Register a page 'mapper' and a notification manager for the given 'fileid'. */

void FilePaging::set(fileid_t fileid, PageMapper *mapper)
{
    FileMapping::iterator entry = _mappers.find(fileid);

    if (entry != _mappers.end())
        return;

    _mappers[fileid] = mapper;

    FileNotification *notifier = new FileNotification;

    _notifiers[fileid] = notifier;
    notifier->attach();
}



/* Convert opening flags to map-compatible paging flags. */

map_flags_t FilePaging::get_flags(flags_t flags)
{
    return flags & (O_WRONLY | O_RDWR) ? L4RE_DS_MAP_FLAG_RW : L4RE_DS_MAP_FLAG_RO;
}



/* Obtain a page mapper for the 'fileid' or register a new one in the
   paging object. */

long FilePaging::get_mapper(FileOpening *opening, const char *path, flags_t flags, fileid_t fileid, PageMapper **mapper)
{
    /* Obtain any registered page mapper. */

    *mapper = get(fileid);

    if (*mapper != NULL)
        return L4_EOK;

    /* Make an accessor and page mapper, registering the mapper. */

    Accessor *accessor;
    long err = opening->make_accessor(path, flags, fileid, &accessor);

    if (err)
        return err;

    *mapper = new PageMapper(accessor, _pages);

    set(fileid, *mapper);

    return L4_EOK;
}

/* Obtain a file-specific notification manager. */

FileNotification *FilePaging::get_notifier(fileid_t file)
{
    FileNotifiers::iterator entry = _notifiers.find(file);

    if (entry != _notifiers.end())
        return _notifiers[file];
    else
        return NULL;
}



/* Return a pager initialised with a page mapper. */

long FilePaging::get_pager(FileOpening *opening, const char *path, flags_t flags, fileid_t fileid, Pager **pager)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Obtain any existing page mapper registered for the file, or make a new
       page mapper. */

    PageMapper *mapper;
    long err = get_mapper(opening, path, flags, fileid, &mapper);

    if (err)
        return err;

    /* Initialise the pager with the mapper and a reference to this object for
       closing the mapper and accessor. */

    *pager = new FilePager(fileid, mapper, get_flags(flags), this);
    return L4_EOK;
}

/* Detach a pager, potentially removing its resources. */

void FilePaging::detach_pager(fileid_t fileid, PageMapper *mapper)
{
    std::lock_guard<std::mutex> guard(_lock);

    if (!mapper->detach())
        remove(fileid, mapper);

    FileNotification *notifier = get_notifier(fileid);

    if ((notifier != NULL) && (!notifier->detach()))
        remove(fileid, notifier);
}

/* Obtain a file-specific notification manager. */

FileNotification *FilePaging::notifier(fileid_t file)
{
    std::lock_guard<std::mutex> guard(_lock);

    return get_notifier(file);
}

// vim: tabstop=4 expandtab shiftwidth=4
